library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity vga is

  port (
    pR     : out std_logic_vector(7 downto 0);
    pG     : out std_logic_vector(7 downto 0);
    pB     : out std_logic_vector(7 downto 0);
    pHSYNC : out std_logic;
    pVSYNC : out std_logic;

    mem_raddr : out std_logic_vector(31 downto 0);
    mem_rdata : in  std_logic_vector(24 downto 0);

    vcount_out : out std_logic_vector(31 downto 0);
    hcount_out : out std_logic_vector(31 downto 0);
    
    clk    : in std_logic;
    reset  : in std_logic;
    debug  : out std_logic_vector(7 downto 0)
    );
  
end vga;

architecture RTL of vga is

  signal hcount : std_logic_vector(31 downto 0);
  signal vcount : std_logic_vector(31 downto 0);

  signal vga_r_out : std_logic_vector(3 downto 0);
  signal vga_g_out : std_logic_vector(3 downto 0);
  signal vga_b_out : std_logic_vector(3 downto 0);

  signal vga_hsync_out : std_logic;
  signal vga_vsync_out : std_logic;

  constant H_ACTIVE_PIXEL_LIMIT : integer := 640;
  constant H_FPORCH_PIXEL_LIMIT : integer := H_ACTIVE_PIXEL_LIMIT + 16;
  constant H_SYNC_PIXEL_LIMIT   : integer := H_FPORCH_PIXEL_LIMIT + 96;
  constant H_BPORCH_PIXEL_LIMIT : integer := H_SYNC_PIXEL_LIMIT + 48;

  constant V_ACTIVE_LINE_LIMIT  : integer := 480;
  constant V_FPORCH_LINE_LIMIT  : integer := V_ACTIVE_LINE_LIMIT + 10;
  constant V_SYNC_LINE_LIMIT    : integer := V_FPORCH_LINE_LIMIT + 2;
  constant V_BPORCH_LINE_LIMIT  : integer := V_SYNC_LINE_LIMIT + 29;

  signal raddr : std_logic_vector(31 downto 0);

begin  -- RTL

  pR <= mem_rdata(23 downto 16) when conv_integer(hcount) < H_ACTIVE_PIXEL_LIMIT and conv_integer(vcount) < V_ACTIVE_LINE_LIMIT else X"00";
  pG <= mem_rdata(15 downto  8) when conv_integer(hcount) < H_ACTIVE_PIXEL_LIMIT and conv_integer(vcount) < V_ACTIVE_LINE_LIMIT else X"00";
  pB <= mem_rdata( 7 downto  0) when conv_integer(hcount) < H_ACTIVE_PIXEL_LIMIT and conv_integer(vcount) < V_ACTIVE_LINE_LIMIT else X"00";
  pHSYNC <= vga_hsync_out;
  pVSYNC <= vga_vsync_out;

  mem_raddr <= raddr;
  vcount_out <= vcount;
  hcount_out <= hcount;

  process (clk, reset)
  begin  -- process
    if reset = '1' then
      hcount <= (others => '0');
      vcount <= (others => '0');
      raddr <= (others => '0');
    elsif clk'event and clk = '1' then  -- rising clock edge
      if (conv_integer(hcount) < H_BPORCH_PIXEL_LIMIT) then
        hcount <= hcount + 1;
      else
        hcount <= (others => '0');
        if (vcount < V_BPORCH_LINE_LIMIT) then
          vcount <= vcount + 1;
        else
          raddr <= (others => '0');
          vcount <= (others => '0');
        end if;
      end if;

      if (conv_integer(hcount) < H_ACTIVE_PIXEL_LIMIT and conv_integer(vcount) < V_ACTIVE_LINE_LIMIT) then -- active data
        raddr <= raddr + 1;
      else
        case (conv_integer(hcount)) is
          when H_ACTIVE_PIXEL_LIMIT => -- front porch
            null;
          when H_FPORCH_PIXEL_LIMIT => -- sync pulse
            vga_hsync_out <= '0';
          when H_SYNC_PIXEL_LIMIT => -- back porch
            vga_hsync_out <= '1';
          when others => null;
        end case;

        case (conv_integer(vcount)) is
          when V_ACTIVE_LINE_LIMIT => -- front porch
            null;
          when V_FPORCH_LINE_LIMIT => -- sync pulse
            vga_vsync_out <= '0';
          when V_SYNC_LINE_LIMIT => -- back porch
            vga_vsync_out <= '1';
          when others => null;
        end case;
      end if;
    end if;
  end process;

end RTL;
